/*
 * Copyright (c) 2001, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

/*
 * @test
 * @bug 4486049
 * @summary toString method fails if size changes in between a call to size
 *           and an attempt to iterate.
 * @author Josh Bloch
 */

import java.util.*;

public class ToString {
    public static void main(String[] args) {
        Map m = new LyingMap();
        if (!m.toString().equals("{}"))
            throw new RuntimeException(m.toString() + "!= {}");

        m.put("x", "1");
        if (!m.toString().equals("{x=1}"))
            throw new RuntimeException(m.toString() + "!= {x=1}");

        m.put("y", "2");
        if (!m.toString().equals("{x=1, y=2}"))
            throw new RuntimeException(m.toString() + "!= {x=1, y=2}");
    }
}

class LyingMap extends LinkedHashMap {
    public int size() {
        return super.size() + 1; // Lies, lies, all lies!
    }
}
